(*
 This structure provides several support functions
 for the LambdaCube structure.

 author: Christian Mueller
*)

local
    open AbstractSyntax
in

datatype mixture = Str of string | E of Exp (* for better error reporting *)

(* c (cutoff) is number of binders passed; variable occurrence x is
local iff x<c *)

fun shift c d (Var (i,s,x))             = Var(i,s,if x<c then x else x+d)
  | shift c d (App(info,e1,e2))         = App(info,shift c d e1, shift c d e2)
  | shift c d (Abstr(info,s,e1,e2))     = 
    Abstr(info,s,shift c d e1, shift (c+1) d e2)

  | shift c d (Pi(info,s,e1,e2))        = 
    Pi(info,s,shift c d e1, shift (c+1) d e2)

  (* additional stuff for integers and boolean *)
  | shift c d (IntNeg (info,e))         = IntNeg(info,shift c d e)
  | shift c d (Plus(info,e1,e2))        = Plus(info,shift c d e1, shift c d e2)
  | shift c d (Minus(info,e1,e2))       = 
    Minus(info,shift c d e1, shift c d e2)

  | shift c d (Mult(info,e1,e2))        = Mult(info,shift c d e1, shift c d e2)
  | shift c d (IntEqual(info,e1,e2))    = 
    IntEqual(info,shift c d e1, shift c d e2)

  | shift c d (IntNotEqual(info,e1,e2)) = 
    IntNotEqual(info,shift c d e1, shift c d e2)

  | shift c d (LessEq(info,e1,e2))      = 
    LessEq(info,shift c d e1, shift c d e2)

  | shift c d (Less(info,e1,e2))        = Less(info,shift c d e1, shift c d e2)
  | shift c d (GreaterEq(info,e1,e2))   = 
    GreaterEq(info,shift c d e1, shift c d e2)

  | shift c d (Greater(info,e1,e2))     = 
    Greater(info,shift c d e1, shift c d e2)

  | shift c d (BoolNeg (info,e))        = BoolNeg(info,shift c d e)
  | shift c d (If(info,e1,e2,e3))       = 
    If(info,shift c d e1, shift c d e2,shift c d e3)       

  (* additional rules for variants *)
  | shift c d (VariantSort (info,xs))   = 
    VariantSort (info, map (fn (s,e) => (s,shift c d e)) xs)

  | shift c d (Variant(info,s,e1,e2))   = 
    Variant (info,s,shift c d e1, shift c d e2)

  | shift c d (Case(info,e,xs))         = 
    Case(info,shift c d e, map (fn (s1,s2,e)=>(s1,s2,shift (c+1) d e)) xs)

  (* additional stuff for records *)
  | shift c d (RecordSort(info,xs))     = 
    RecordSort (info, map (fn (s,e) => (s,shift c d e)) xs)

  | shift c d (Record(info,xs))         = 
    Record (info, map (fn (s,e) => (s,shift c d e)) xs)

  | shift c d (Proj(info,e,s))          = Proj(info,shift c d e,s)
  (* additional rules for iso-recursive types *)
  | shift c d (Fold(info,T,e))          = Fold(info,shift c d T, shift c d e)
  | shift c d (Unfold(info,T,e))        = Unfold(info,shift c d T, shift c d e)
  | shift c d (RecursiveSort(info,s,e)) = 
    RecursiveSort(info,s, shift (c+1) d e) 

  (* Ascription *)
  | shift c d (Ascribe(info,e1,e2))     = 
    Ascribe(info,shift c d e1, shift c d e2)
    
  (* built-in fixpoint operator *)
  | shift c d (Fix(info,e))             = Fix(info, shift c d e)

  (* references *)
  | shift c d (Ref(i,e))                = Ref(i,shift c d e)
  | shift c d (Deref(i,e))              = Deref(i,shift c d e)
  | shift c d (Assign(i,e1,e2))         = Assign(i,shift c d e1,shift c d e2)
  | shift c d (RefType(i,e))            = RefType(i,shift c d e)

  (* let expression *)
  | shift c d (Let(i,s,e1,e2))         = Let(i,s,shift c d e1,shift (c+1) d e2)
  | shift _ _ const                    = const


(* substitutes e for y if x=y+c; x<c needed to protect against misuse with 
   negative cutoff *)

fun subst c (Var (info,s,x))  y e         = 
    if x<c orelse x<>y+c then Var (info,s,x) else shift 0 c e

  | subst c (App(info,e1,e2)) y e         = 
	App(info,subst c e1 y e, subst c e2 y e)

  | subst c (Abstr(info,s,e1,e2)) y e     = 
	Abstr(info,s,subst c e1 y e, subst (c+1) e2 y e)

  | subst c (Pi(info,s,e1,e2)) y e        = 
	Pi(info,s,subst c e1 y e, subst (c+1) e2 y e)

  (* additional rules for integers and boolean *)
  | subst c (IntNeg (info,e1)) y e        = IntNeg(info,subst c e1 y e)
  | subst c (Plus(info,e1,e2)) y e        = 
	Plus(info,subst c e1 y e, subst c e2 y e)

  | subst c (Minus(info,e1,e2)) y e       = 
	Minus(info,subst c e1 y e, subst c e2 y e)

  | subst c (Mult(info,e1,e2)) y e        = 
	Mult(info,subst c e1 y e, subst c e2 y e)

  | subst c (IntEqual(info,e1,e2)) y e    = 
	IntEqual(info,subst c e1 y e, subst c e2 y e)

  | subst c (IntNotEqual(info,e1,e2)) y e = 
	IntNotEqual(info,subst c e1 y e, subst c e2 y e)
	
  | subst c (LessEq(info,e1,e2)) y e      = 
	LessEq(info,subst c e1 y e, subst c e2 y e)

  | subst c (Less(info,e1,e2)) y e        = 
	Less(info,subst c e1 y e, subst c e2 y e)

  | subst c (GreaterEq(info,e1,e2)) y e   = 
	GreaterEq(info,subst c e1 y e, subst c e2 y e)

  | subst c (Greater(info,e1,e2)) y e     = 
	Greater(info,subst c e1 y e, subst c e2 y e)

  | subst c (BoolNeg (info,e1)) y e       = BoolNeg(info,subst c e1 y e)
  | subst c (If(info,e1,e2,e3)) y e       = 
	If(info,subst c e1 y e, subst c e2 y e, subst c e3 y e)       

  (* additional rules for variants *)
  | subst c (VariantSort (info,xs)) y e   = 
	VariantSort (info, map (fn (s,e') => (s,subst c e' y e)) xs)

  | subst c (Variant(info,s,e1,e2)) y e   = 
	Variant (info,s,subst c e1 y e, subst c e2 y e)

  | subst c (Case(info,e1,xs))    y e     = 
	Case(info,subst c e1 y e, 
	     map (fn (s1,s2,e')=>(s1,s2,subst (c+1) e' y e)) xs)

  (* additional rules for records *)
  | subst c (RecordSort(info,xs)) y e     = 
	RecordSort (info, map (fn (s,e') => (s,subst c e' y e)) xs)

  | subst c (Record(info,xs)) y e         = 
	Record (info, map (fn (s,e') => (s,subst c e' y e)) xs)

  | subst c (Proj(info,e',s)) y e         = Proj(info, subst c e' y e,s)
  (* additional rules for iso-recursive types *)
  | subst c (Fold(info,T,e')) y e         = 
	Fold(info,subst c T y e, subst c e' y e)

  | subst c (Unfold(info,T,e')) y e       = 
	Unfold(info,subst c T y e, subst c e' y e)

  | subst c (RecursiveSort(info,s,T)) y e = 
	RecursiveSort(info,s, subst (c+1) T y e)

  (* Ascription *)
  | subst c (Ascribe(info,e1,e2)) y e     = 
	Ascribe(info,subst c e1 y e, subst c e2 y e)

  (* built-in fixpoint operator *)
  | subst c (Fix(info,e')) y e            = Fix(info, subst c e' y e)
  (* references *)
  | subst c (Ref(i,e1)) y e               = Ref(i,subst c e1 y e)
  | subst c (Deref(i,e1)) y e             = Deref(i,subst c e1 y e)
  | subst c (RefType(i,e1)) y e           = RefType(i,subst c e1 y e)
  | subst c (Assign(i,e1,e2)) y e         = 
        Assign(i,subst c e1 y e, subst c e2 y e)

  (* let expression *)
  | subst c (Let(i,s,e1,e2)) y e          = 
        Let(i,s,subst c e1 y e,subst (c+1) e2 y e)

  | subst _ const       _ _               = const



(* checks whether variable with index c occurs in the expression *)

fun occurs (Var (info,s,x)) c          = x=c
  | occurs (App(info,e1,e2)) c         = occurs e1 c orelse occurs e2 c
  | occurs (Abstr(info,s,e1,e2)) c     = occurs e1 c orelse occurs e2 (c+1)
  | occurs (Pi(info,s,e1,e2)) c        = occurs e1 c orelse occurs e2 (c+1)
  (* additional rules for integers and boolean *)
  | occurs (IntNeg (info,e1)) c        = occurs e1 c
  | occurs (Plus(info,e1,e2)) c        = occurs e1 c orelse occurs e2 c
  | occurs (Minus(info,e1,e2)) c       = occurs e1 c orelse occurs e2 c
  | occurs (Mult(info,e1,e2)) c        = occurs e1 c orelse occurs e2 c
  | occurs (IntEqual(info,e1,e2)) c    = occurs e1 c orelse occurs e2 c
  | occurs (IntNotEqual(info,e1,e2)) c = occurs e1 c orelse occurs e2 c
  | occurs (LessEq(info,e1,e2)) c      = occurs e1 c orelse occurs e2 c
  | occurs (Less(info,e1,e2)) c        = occurs e1 c orelse occurs e2 c
  | occurs (GreaterEq(info,e1,e2)) c   = occurs e1 c orelse occurs e2 c
  | occurs (Greater(info,e1,e2)) c     = occurs e1 c orelse occurs e2 c
  | occurs (BoolNeg (info,e1)) c       = occurs e1 c
  | occurs (If(info,e1,e2,e3)) c       = (occurs e1 c orelse occurs e2 c 
					  orelse occurs e3 c)
  (* additional rules for variants *)
  | occurs (VariantSort (info,xs)) c   = 
    List.exists (fn (s,e') => occurs e' c) xs

  | occurs (Variant(info,s,e1,e2)) c   = occurs e1 c orelse occurs e2 c
  | occurs (Case(info,e1,xs))    c     = 
    occurs e1 c orelse List.exists (fn (s1,s2,e')=> occurs e' (c+1)) xs

  (* additional rules for records *)
  | occurs (RecordSort(info,xs)) c     = 
    List.exists (fn (s,e') => occurs e' c) xs

  | occurs (Record(info,xs)) c         = 
    List.exists (fn (s,e') => occurs e' c) xs
  | occurs (Proj(info,e',s)) c         = occurs e' c
  (* additional rules for iso-recursive types *)
  | occurs (Fold(info,T,e')) c         = occurs T c orelse occurs e' c
  | occurs (Unfold(info,T,e')) c       = occurs T c orelse occurs e' c
  | occurs (RecursiveSort(info,s,T)) c = occurs T (c+1)
  (* Ascription *)
  | occurs (Ascribe(info,e1,e2)) c     = occurs e1 c orelse occurs e2 c
  (* built-in fixpoint operator *)
  | occurs (Fix(info,e')) c            = occurs e' c
  (* references *)
  | occurs (Ref(_,e)) c                = occurs e c
  | occurs (Deref(_,e)) c              = occurs e c
  | occurs (RefType(_,e)) c            = occurs e c
  | occurs (Assign(_,e1,e2)) c         = occurs e1 c orelse occurs e2 c  
  (* let expression *)
  | occurs (Let(_,_,e1,e2)) c          = occurs e1 (c+1) orelse occurs e2 c
  | occurs _  _                        = false 

end
