(* 
   Lambda Cube 

   Core module of the interpreter
   
   author: Christian Mueller   
*)

local
    open AbstractSyntax LCSupport Info Support
in

exception Error of string
exception ErrorWInfo of info * (mixture list)

(* beta-reduction that preserves de Bruijn numbers *)

fun beta e e' = shift 0 ~1 (subst 0 e 0 (shift 0 1 e'))

(* evaluation of expressions *)

exception NoFurtherEval

fun isval e = 
    case e of 
	Abstr _ => true
      | Var _ => true
      | Int _ => true
      | True _ => true
      | False _ => true
      | Unit _ => true
      | Variant (_,_,e,_) => isval e
      | Record (_,xs) => List.all (fn (_,e) => isval e) xs
      | Fold(_,_,e) => isval e
      | Loc _ => true
      (* type "values": no recursion because of lazy evaluation of types *)
      | IntType _ => true
      | BoolType _ => true
      | UnitType _ => true
      | RefType _ => true
      | Pi _ => true
      | RecursiveSort _ => true
      | VariantSort _ => true
      | RecordSort _ => true
      | _ => false

fun onestep_eval(App(_ , v1 as Abstr(_,_,_,e1),e2)) = 
    if isval e2 then beta e1 e2 
    else App(emptyInfo,v1,onestep_eval e2)

  | onestep_eval(App(_,e1,e2)) = App(emptyInfo,onestep_eval e1,e2)
  (* integer and boolean operations *)
  | onestep_eval(IntNeg (info,Int (_,i)))  = Int (info,~i)
  | onestep_eval(IntNeg (info,e))          = IntNeg (info,onestep_eval e)

  | onestep_eval(Plus(info,Int (_,x), Int (_,y)))  = Int (emptyInfo,x+y)
  | onestep_eval(Plus(info,Int (i,x), e)) = 
	Plus(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(Plus(info,e1,e2)) = Plus(emptyInfo,onestep_eval e1, e2)  

  | onestep_eval(Minus(info,Int (_,x), Int (_,y))) = Int (emptyInfo,x-y)
  | onestep_eval(Minus(info,Int (i,x), e)) = 
	Minus(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(Minus(info,e1,e2)) = Minus(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(Mult(info,Int (_,x), Int (_,y)))  = Int (emptyInfo,x*y)
  | onestep_eval(Mult(info,Int (i,x), e)) = 
	Mult(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(Mult(info,e1,e2)) = Mult(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(BoolNeg (info,(True emptyInfo))) = False emptyInfo
  | onestep_eval(BoolNeg (info,False emptyInfo))  = True emptyInfo
  | onestep_eval(BoolNeg (info,e))    = BoolNeg(emptyInfo,onestep_eval e)

  | onestep_eval(IntEqual(info,Int (_,x), Int (_,y))) = 
	if x=y then True emptyInfo else False emptyInfo
  | onestep_eval(IntEqual(info,Int (i,x), e)) = 
	    IntEqual(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(IntEqual(info,e1,e2)) = 
        IntEqual(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(IntNotEqual(info,Int (_,x), Int (_,y))) = 
	    if x<>y then True emptyInfo else False emptyInfo
  | onestep_eval(IntNotEqual(info,Int (i,x), e)) = 
		IntNotEqual(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(IntNotEqual(info,e1,e2))                = 
		IntNotEqual(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(LessEq(info,Int (_,x), Int (_,y)))      = 
		if x<=y then True emptyInfo else False emptyInfo
  | onestep_eval(LessEq(info,Int (i,x), e))              = 
		    LessEq(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(LessEq(info,e1,e2))                     = 
		    LessEq(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(Less(info,Int (_,x), Int (_,y)))        = 
		    if x<y then True emptyInfo else False emptyInfo
  | onestep_eval(Less(info,Int (i,x), e))= 
         Less(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(Less(info,e1,e2)) = Less(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(GreaterEq(info,Int (_,x), Int (_,y)))   = 
			if x>=y then True emptyInfo else False emptyInfo
  | onestep_eval(GreaterEq(info,Int (i,x), e)) = 
			    GreaterEq(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(GreaterEq(info,e1,e2)) = 
         GreaterEq(emptyInfo,onestep_eval e1, e2)

  | onestep_eval(Greater(info,Int (_,x), Int (_,y)))     = 
	  if x>y then True emptyInfo else False emptyInfo
  | onestep_eval(Greater(info,Int (i,x), e)) = 
	      Greater(emptyInfo,Int (i,x), onestep_eval e)
  | onestep_eval(Greater(info,e1,e2)) = 
         Greater(emptyInfo,onestep_eval e1, e2)
  | onestep_eval(If(_,True emptyInfo,e1,_))  = e1
  | onestep_eval(If(_,False emptyInfo,_,e2)) = e2
  | onestep_eval(If(_,e1,e2,e3))   = If(emptyInfo,onestep_eval e1,e2,e3)
  (* case and variant *)
  | onestep_eval(c as Case(_,Variant(info,l,e',T),xs)) = 
    (if isval e' then   
	 beta (case (List.find (fn (s,_,_) => s=l) xs) of 
		   NONE  => raise Error "onestep_eval, case: wrong label" 
		 | SOME (_,_,e) => e ) e' 
     else
	 Case(emptyInfo,Variant(info,l,onestep_eval e',T),xs) )
  | onestep_eval(Case(_,e,xs))          = Case(emptyInfo,onestep_eval e,xs)
  | onestep_eval(Variant(info,s,e1,e2)) = 
         Variant(emptyInfo,s,onestep_eval e1, e2)
  (* record *)
  | onestep_eval(r as Record(info,xs)) = 
	 let 
	     fun rappend nil ys = ys
	       | rappend (x::xr) ys = rappend xr (x::ys)

	     fun eval_comp nil ys = rev ys
	       | eval_comp ((l,t)::xr) ys = 
		 if isval t then eval_comp xr ((l,t)::ys) 
		 else rappend ((l,onestep_eval t)::ys)  xr
	 in
	     if isval r then raise NoFurtherEval
	     else Record(info,eval_comp xs nil)
	 end
  | onestep_eval(Proj(info,r as Record(_,xs),s)) = 
	 (if isval r then 
	      case List.find (fn (l,_)=> s=l) xs of
		  NONE => raise Error ("eval: no record field \""^s^"\" found")
		| SOME (_,v) => v
	  else
	      Proj(emptyInfo,onestep_eval r,s))
  | onestep_eval(Proj(info,e,s)) = Proj(emptyInfo,onestep_eval e,s)
  (* iso-recursive types *)
  | onestep_eval(Unfold(_,_,Fold(_,_,e))) = e 
  | onestep_eval(Fold(info,T,e))    = Fold(emptyInfo,T,onestep_eval e)
  | onestep_eval(Unfold(info,T,e))  = Unfold(emptyInfo,T, onestep_eval e)
  (* ascription *)
  | onestep_eval(Ascribe(info,e,T)) = 
	      if isval e then e else Ascribe(emptyInfo,onestep_eval e,T)
  (* built-in fixpoint operator *)
  | onestep_eval(f as Fix(info,Abstr(_,_,_,e)))  = beta e f
  | onestep_eval(Fix(info,e))                    = Fix(info, onestep_eval e)
  (* references based on a global store *)
  | onestep_eval(Ref(i,e)) = if isval e then Loc(emptyInfo, Store.newref e)
			     else Ref(i,onestep_eval e)
  | onestep_eval(Deref(i,e))= 
    (case e of 
	 Loc (_,l) => Store.read l
       | _ => if isval e 
	      then raise ErrorWInfo (i,[Str "deref: cannot evaluate ",
					E e,
					Str " to a location"])
	      else Deref(i,onestep_eval e))

  | onestep_eval(Assign(i,e1,e2)) = 
    if isval e1 andalso isval e2 then
	(case e1 of
	     Loc (_,l) => (Store.assign l e2; Unit emptyInfo)
	   | _ => raise ErrorWInfo(i,[Str "assign: cannot evaluate ",
				      E e1,
				      Str " to a location"]))
    else if isval e1 then Assign(i,e1,onestep_eval e2)
         else Assign(i,onestep_eval e1,e2)         

  (* let expression *)
  | onestep_eval (Let(_,s,e1,e2)) = if isval e1 then beta e2 e1
				    else Let(emptyInfo,s,onestep_eval e1,e2)
	 
  | onestep_eval e                  = raise NoFurtherEval

(* evaluate until no more steps are possible *)

fun eval e = eval(onestep_eval e) handle NoFurtherEval => e

(* type environment *)

type environment = Exp list

fun get g x s = 
    shift 0 (x+1) (List.nth(g,x) 
    handle Subscript => raise Error("unbound variable "^s^":"^Int.toString x))

(* check for equivalence of types modulo weak beta-reduction *)

fun equiv (s,t) =    
    (s = t orelse
     case (eval s, eval t) of	
	(* this is the core part needed by S,F,Fw and CC *)
        (UnitType _,UnitType _)    => true
      | (IntType _, IntType _)     => true
      | (BoolType _, BoolType _)   => true
      | (RefType (_,e1),RefType(_,e2)) => equiv (e1,e2)
      | (Var (_,_,x), Var (_,_,y)) => x=y 
      | (Pi(_,_,e11,e12), Pi(_,_,e21,e22)) => 
	    equiv (e11,e21) andalso equiv (e12,e22)

      | (VariantSort (_,xs), VariantSort (_,ys))      => checkLists xs ys
      | (RecordSort(_,xs),RecordSort(_,ys))           => checkLists xs ys
      | (RecursiveSort(_,_,T1),RecursiveSort(_,_,T2)) => equiv (T1,T2)       
        (* needed by F,Fw,CC to allow application of types;
	   for IF and CASE use sigglex to prevent dependent types 
           in S,F and Fw *)
      |	(Star _, Star _)           => true 
	(* needed by Fw and CC for type operators where one or both side(s) 
	   of an application might be abstracted *)
      | (Abstr(_,_,e11,e12),Abstr(_,_,e21,e22)) => 
	    equiv (e11,e21) andalso equiv (e12,e22)

      | (App(_,e11,e12),App(_,e21,e22)) => 
	    equiv (e11,e21) andalso equiv (e12,e22) 

        (* the following is only needed for CC and results in full equivalence
	   check of arbitrary expressions*)
      | (If(_,e1,e2,e3),If(_,e1',e2',e3')) => 
	    equiv(e1,e1') andalso equiv(e2,e2') andalso equiv(e3,e3')

      | (Plus(_,e1,e2),Plus(_,e1',e2')) => equiv(e1,e1') andalso equiv(e2,e2')
      | (Minus(_,e1,e2),Minus(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (Mult(_,e1,e2),Mult(_,e1',e2')) => equiv(e1,e1') andalso equiv(e2,e2')
      | (IntEqual(_,e1,e2),IntEqual(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (IntNotEqual(_,e1,e2),IntNotEqual(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (Less(_,e1,e2),Less(_,e1',e2')) => equiv(e1,e1') andalso equiv(e2,e2')
      | (LessEq(_,e1,e2),LessEq(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (Greater(_,e1,e2),Greater(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (GreaterEq(_,e1,e2),GreaterEq(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (IntNeg (_,e1), IntNeg (_,e2)) => equiv(e1,e2)
      | (BoolNeg (_,e1), BoolNeg (_,e2)) => equiv(e1,e2)	    
      | (Int (_,x),Int (_,y)) => x=y
      | (True _,True _) => true
      | (False _, False _) => true
      | (Unit _, Unit _) => true
      | (Fold(_,e1,e2),Fold(_,e1',e2')) => equiv(e1,e1') andalso equiv(e2,e2')
      | (Unfold(_,e1,e2),Unfold(_,e1',e2')) => 
	    equiv(e1,e1') andalso equiv(e2,e2')

      | (Proj(_,e1,l1),Proj(_,e2,l2)) => l1=l2 andalso equiv(e1,e2)
      | (Record(_,xs),Record(_,ys)) => checkLists xs ys
      | (Case(_,e,xs),Case(_,e',ys)) => equiv(e,e') andalso checkCase xs ys
      | (Variant(_,s,e1,e2),Variant(_,s',e1',e2')) => 
	    s=s' andalso equiv(e1,e1') andalso equiv(e2,e2')

      | (Fix(_,e),Fix(_,e')) => equiv(e,e')
      | (Ref(_,e1),Ref(_,e2)) => equiv (e1,e2)
      | (Deref(_,e1),Deref(_,e2)) => equiv (e1,e2)
      | (Loc(_,l1),Loc(_,l2)) => l1=l2
      | (Assign(_,e11,e12),Assign(_,e21,e22)) => 
	   equiv (e11,e21) andalso equiv (e12,e22)
	
      | (e1,e2) => false)

and checkLists xs ys = 
    let
	(* variant and record types must be sorted for equivalence check *)
	val sort = Listsort.sort 
	    (fn ((s1,_),(s2,_)) => String.compare (s1,s2)) 

	fun c (nil,nil) = true 
	  | c ((l1,e1)::xr,(l2,e2)::yr) = 
	    l1=l2 andalso equiv (e1,e2) andalso c (xr,yr) 
	  | c _ = false	     
    in 
	c (sort xs, sort ys) 
    end

and checkCase xs ys = 
    let
	(* case fields must be sorted for equivalence check *)
	val sort = Listsort.sort 
	    (fn ((s1,_,_),(s2,_,_)) => String.compare (s1,s2))

	fun c (nil,nil) = true
	  | c ((l1,_,e1)::xr,(l2,_,e2)::yr) = 
	    l1=l2 andalso equiv (e1,e2) andalso c (xr,yr)
	  | c _ = false	    
    in
	c (sort xs, sort ys)
    end


(* sup and inf *)

val sort = Listsort.sort (fn ((s1,_),(s2,_)) => String.compare (s1,s2)) 

fun sup (s,t) = 
    if equiv (s,t) then s 
    else (case (s,t) of 
	      (Pi(_,_,e11,e12), Pi(_,_,e21,e22))  => 
		  Pi(emptyInfo,"x",inf (e11,e21), sup(e12,e22))

	    | (RecordSort(_,xs),RecordSort(_,ys)) => 
	      let
		  val xs = suprec sup (sort xs, sort ys) nil
	      in 
		  if null xs then raise Error "incompatible record types" 
		  else RecordSort(emptyInfo,xs)
	      end
	    | (VariantSort (_,xs), VariantSort (_,ys)) => 
	      let 
		  val xs = infrec sup (sort xs, sort ys) nil
	      in 
		  if null xs then raise Error "incompatible variant types" 
		  else VariantSort(emptyInfo,xs)
	      end	
	    | (e1,e2) => raise ErrorWInfo (emptyInfo,
					   [E e1,Str " and ", E e2, 
					    Str " do not have a supremum"]))

and suprec f (nil,_) acc = acc
  | suprec f (_,nil) acc = acc    
  | suprec f ((l1,e1)::xr,(l2,e2)::yr) acc = 
    (case String.compare(l1,l2) of 
	 EQUAL   => suprec f (xr,yr) ((l1,f (e1,e2))::acc)
       | LESS    => suprec f (xr,(l2,e2)::yr) acc
       | GREATER => suprec f ((l1,e1)::xr,yr) acc)

and inf (s,t) = 
    if equiv (s,t) then s
    else (case (s,t) of
	      (Pi(_,_,e11,e12), Pi(_,_,e21,e22))  => 
		  Pi(emptyInfo,"x",sup (e11,e21), inf (e12,e22))

	    | (RecordSort(_,xs),RecordSort(_,ys)) => 
		  RecordSort(emptyInfo,infrec inf (sort xs, sort ys) nil)

	    | (VariantSort (_,xs), VariantSort (_,ys)) => 
		  VariantSort(emptyInfo, suprec inf (sort xs, sort ys) nil)

	    | (e1,e2) => raise ErrorWInfo (emptyInfo,
					   [E e1,Str " and ", E e2, 
					    Str " are not yet ", 
					    Str "supported by inf"]))

and infrec f (nil,yr) acc = yr @ rev acc
  | infrec f (xr,nil) acc = rev acc @ xr
  | infrec f ((l1,e1)::xr,(l2,e2)::yr) acc = 
    (case String.compare(l1,l2) of
	 EQUAL   => infrec f (xr,yr) ((l1,f (e1,e2))::acc)
       | LESS    => infrec f (xr,(l2,e2)::yr) ((l1,e1)::acc)
       | GREATER => infrec f ((l1,e1)::xr,yr) ((l2,e2)::acc))
	 
(* subtyping *)

fun subtype g (s,t) = 
    let
	val sort = Listsort.sort (fn ((s1,_),(s2,_)) => String.compare (s1,s2))
	    
	fun sub' (g1,g2) (s,t) = 
	    (case (eval s, eval t) of
		 (* needed for subtyping in S *)
		 (UnitType _,UnitType _)    => true 
	       | (IntType _, IntType _)     => true 
	       | (BoolType _, BoolType _)   => true 
	       | (v1 as Var (_,_,x), v2 as Var (_,_,y)) => 
		     (x=y andalso 
		      (case List.nth(g1,x) of 
			   Var(_,_,0) => true | _ => false) 
			   handle Subscript => false)

	       | (Pi(_,_,e11,e12), Pi(_,_,e21,e22)) => 
		     sub' (g2,g1) (e21,e11) 
		     andalso sub' (e21::g1,e21::g2) (e12,e22)	

	       | (VariantSort (_,xs), VariantSort (_,ys)) => 
		     subvar (g1,g2) (sort xs, sort ys)

	       | (RecordSort(_,xs),RecordSort(_,ys)) => 
		     subrec (g1,g2) (sort xs, sort ys)

	       | (RecursiveSort(_,_,T1),RecursiveSort(_,_,T2)) => 
		     equiv (T1,T2) orelse 
		     sub' (Var(emptyInfo,"",0)::g1,Box emptyInfo::g2) (T1,T2)
		     
	       (* Ref types must be invariant !!! *)
	       | (RefType (_,e1), RefType (_,e2)) => equiv (e1,e2)

	       (* neede for subtyping in F *)
	       | (Star _, Star _)           => true  
	       (* needed for subtyping in Fw *)
	       | (Abstr(_,_,e11,e12),Abstr(_,_,e21,e22))  => 
		     equiv(e21,e11) andalso sub' (e21::g1,e11::g2) (e12,e22)

	       | (App(_,e11,e12),App(_,e21,e22)) => 
		     sub' (g1,g2) (e11,e21) andalso equiv (e12,e22)
	       (* no extra rules for CC *)
	       | (e1,e2) => false)
	 
	and subrec (g1,g2) (_,nil) = true	   
	  | subrec (g1,g2) ((l1,e1)::xr,(l2,e2)::yr) = 
	    if l1=l2 then (sub' (g1,g2) (e1,e2) andalso subrec (g1,g2) (xr,yr))
	    else  subrec (g1,g2) (xr,(l2,e2)::yr)
	  | subrec _ _ = false
	    
	and subvar (g1,g2) (nil,_) = true	  
	  | subvar (g1,g2) ((l1,e1)::xr,(l2,e2)::yr) = 
	    if l1=l2 then 
		(sub' (g1,g2) (e1,e2) andalso subvar (g1,g2) (xr,yr)) 
	    else subvar (g1,g2) ((l1,e1)::xr,yr)  

	  | subvar _ _ = false
    in
	(* as the subtyping relation is not generalized for subtyping of all 
	   features in CC, equiv is used first to ensure that all expressions
	   that are typable in non-subtyping mode are still typable *)

	equiv (s,t) orelse sub' (g,g) (s,t)
    end  


(* adjust relation to global settings *)

fun compare g (e1,e2) = if !subtyping then subtype g (e1,e2) else equiv (e1,e2)

(* sets up the interpreter typing mode *)

fun S_squiggle (Star _) (Star _) = Star emptyInfo
  | S_squiggle s t = 
    raise ErrorWInfo (emptyInfo,
		      [E s, Str "->", E t, 
		       Str " is an illegal quantification in ", 
		       Str (case !typingmode of 
				S=>"S"|F=>"F"|Fw=>"Fw"
			      |CC=>"CC"|_=>"internal error")])
		     

fun F_squiggle (Box _) (Star _) = Star emptyInfo
  | F_squiggle s t = S_squiggle s t

fun Fw_squiggle (Box _) (Box _)  = Box emptyInfo
  | Fw_squiggle s t = F_squiggle s t

fun CC_squiggle (Star _) (Box _) = Box emptyInfo
  | CC_squiggle s t = Fw_squiggle s t 

val squiggle = ref (fn _ => fn _ => raise Fail "no type mode set")

fun setTypingMode S  = squiggle := S_squiggle 
  | setTypingMode F  = squiggle := F_squiggle
  | setTypingMode Fw = squiggle := Fw_squiggle 
  | setTypingMode CC = squiggle := CC_squiggle 
  | setTypingMode _  = squiggle := 
    (fn _ => fn _ => raise Fail "LC: no type mode set")

(* type checking procedure *)
  
fun check e = check' nil e

and check' g (Var(_,s,x)) = get g x s
  | check' g (App(info,e,e')) = 
    (let 
	 val T  = eval(check' g e)
	 val T' = check' g e' 
     in 
	 case T of 
	     Pi(_,_,t',t) => 
		 if compare g (T',t') then beta t e'
		 else raise ErrorWInfo (info,
					[Str "argument of type ", 
					 E T', 
					 Str " cannot be applied to a",
					 Str " function ", E e,
					 Str " of type ",E T])
	   | t => raise ErrorWInfo (info,
				    [E t, Str " is not a procedure type"]) 
     end)

  | check' g (Abstr(_,s,t,e))  =  
    let 
	val _ = check' g t
	val t'= Pi(emptyInfo,s,t,check' (t::g) e)
    in 
	check' g t' ; t' 
    end
  | check' g (Pi(_,_,t,t'))  = (!squiggle) (check' g t) (check' (t::g) t')
  | check' g (Star _)        = Box emptyInfo
  | check' g (Box _)         = raise Error "Box found"
  (* unit check's *)
  | check' _ (UnitType _) = Star emptyInfo
  | check' _ (Unit _)     = UnitType emptyInfo
  (* additional check's for integers and boolean *)
  | check' g (IntType _)            = Star emptyInfo
  | check' _ (BoolType _)           = Star emptyInfo
  | check' _ (Int _)                = IntType emptyInfo
  | check' _ (True _)               = BoolType emptyInfo  
  | check' _ (False _)              = BoolType emptyInfo
  | check' g (BoolNeg (_,e))        = 
    if compare g (check' g e,BoolType emptyInfo) then (BoolType emptyInfo)
    else raise Error "op not: boolean expression expected"

  | check' g (IntNeg (_,e))         = 
    if compare g (check' g e,IntType emptyInfo) then (IntType emptyInfo)
    else raise Error "op~: integer expression expected"

  | check' g (Plus (i,e1,e2)) = 
	checkEq g (e1,e2) (IntType emptyInfo) (IntType emptyInfo)  i "op+"

  | check' g (Minus(i,e1,e2))       = 
	checkEq g (e1,e2) (IntType emptyInfo) (IntType emptyInfo)  i "op-"

  | check' g (Mult(i,e1,e2))        = 
	checkEq g (e1,e2) (IntType emptyInfo) (IntType emptyInfo)  i "op*"

  | check' g (IntEqual(i,e1,e2))    = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op="

  | check' g (IntNotEqual(i,e1,e2)) = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op="

  | check' g (LessEq(i,e1,e2))      = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op<="

  | check' g (Less(i,e1,e2))        = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op="

  | check' g (GreaterEq(i,e1,e2))   = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op>="

  | check' g (Greater(i,e1,e2))     = 
	checkEq g (e1,e2) (IntType emptyInfo) (BoolType emptyInfo) i "op>"

  | check' g (i as If(info,e1,e2,e3))  = checkIf g i
  (* additional rules for variant *)
  | check' g (v as Variant _ )    = checkVariant g v
  | check' g (VariantSort (info,xs)) = 
        checkField g xs info "wrong kinded variant sort"

  | check' g (c as Case _)        = checkCase g c
  (* additional rules for records *)
  | check' g (Proj(_,e,s))      = 
	(case eval(check' g e) of
	     RecordSort (_,xs) => 
		 (case List.find (fn (l,_)=> s=l) xs of
		      NONE => raise Error ("no record field \""^s^"\" found")
		    | SOME (_,T) => T)
	   | _ =>  raise Error "wrong build projection" )

  | check' g (Record(_,xs))     = 
	 let 
	     val T = RecordSort(emptyInfo, map (fn (l,e) => (l,check' g e)) xs)
	 in
	     check' g T; T
	 end

  | check' g (RecordSort(i,xs)) = checkField g xs i "wrong kinded record sort"
  (* additional rules for iso-recursive types *)
  | check' g (Fold(i,T,e))   = 
	 (case (check' g T; eval T) of
	      RecursiveSort(_,x,T1) => 
		  let 
		      val (t,TX) = (check' g e,beta T1 T)
		  in  
		      if compare g (t,TX)  then T
		      else raise ErrorWInfo(i,[Str "expected ",E TX,
					       Str " but got ",E t, 
					       Str " in fold expression"])
		  end
	    | _ => raise Error "wrong type annotation in fold exrepssion")

  | check' g (Unfold(i,T,e)) = 
	  (case (check' g T; eval T) of
	       RecursiveSort(_,x,T1) => 
		   let 
		       val T' = check' g e
		   in 
		       if compare g (T', T) then beta T1 T
		       else raise ErrorWInfo (i,[Str "expected ", E T, 
						 Str " but got ",E T',
						 Str " in unfold expression"])
		   end
	     | _ => raise ErrorWInfo(i,[Str "wrong type annotation in",
					Str " unfold expression"]))

  | check' g (RecursiveSort(i,_,T)) = 
	   if compare g (check' ((Star emptyInfo)::g) T, Star emptyInfo) 
	   then Star emptyInfo
	   else raise ErrorWInfo(i,[Str"wrong built recursive type"])
		    
  (* additional rule for ascription *)
  | check' g (Ascribe(i,e,T')) = 
	  let 
	      val T = check' g e 
	  in 
	      if compare g (T,T') then T'  
	      else raise ErrorWInfo (i,[Str "ascription is ",E T', 
					Str ", but expression has type ", 
					E T]) 
	  end

  (* built-in fixpoint operator *)
  | check' g (Fix(info,e)) = 
	  let 
	      val s = check' g e  		      
	  in 
	      if compare g (Star emptyInfo, check' g s) then
		  (case eval s of 
		       Pi (_,_,e1,e2)=>  
		       if compare g (e2,e1) then e1 
		       else raise ErrorWInfo (info,[Str "fix expexts a type ", 
						    Str "T->T and got ",E s])
		     | _ => raise ErrorWInfo (info,[Str "fix expexts a type ",
						    Str "T->T and got ",E s]))
	      else raise ErrorWInfo(info,[Str "fix only accepts terms,",
					  Str " but got ", E e])	  
	  end

  (* references *)
  | check' g (Ref(info,e)) = 
    let
	val T = RefType(emptyInfo,check' g e)
    in
	check' g T; T
    end
  | check' g (Deref(info,e)) = 
    (case check' g e of 
	 RefType (_,e) => e
       | t => raise ErrorWInfo (info,
				[Str "! expects a Ref Type and got ", E t]))
  | check' g (Assign(i,e1,e2)) = 
    let 
	val (t1,t2) = (check' g e1, check' g e2)
    in
	case t1 of
	    RefType (_,t1') => 
	    if compare g (t1',t2) 
	       andalso compare g (t2,t1') (* in case of subtyping *)
	    then UnitType emptyInfo
	    else raise ErrorWInfo (i,[Str "distinct types in assignment, ",
				      Str "location of type ",E t1, 
				      Str ", rhs of type ", E t2])
			     
	  | _ => raise ErrorWInfo (i,[Str "assignment Ref Type expected ",
				      Str " and got ", E t1])
    end
  | check' g (RefType(i,e)) = 
    if compare g (check' g e,Star emptyInfo) then Star emptyInfo
    else raise ErrorWInfo(i,[Str "no higher ordered ref type allowed"])

  (* this check is needed as we allow step-by-step evaluation *)
  | check' g (Loc(_,l)) = 
    let
	val T = RefType(emptyInfo, check' g (Store.read l))
    in
	check' g T; T
    end

  (* 
     let expression 
	 The type-checker translates the let expression into an application
         and checks this application.
   *)
  | check' g (Let(info,_,e1,e2)) =    
    let
	val T1 = check' g e1
    in
	check' g (App(emptyInfo,Abstr(emptyInfo,"",T1,e2),e1))
    end
				 				      
(* some help functions *)
				
and checkEq g (e1,e2) T T' info error = 
    let
	val T1 = check' g e1
	val T2 = check' g e2
    in
	if compare g(T1,T) andalso compare g (T2,T) then T'
	else raise ErrorWInfo (info,
			       [Str (error^": expexted twice type "), 
				E T, Str " but got ",E T1, Str " and ", E T2])
    end

and checkIf g (If(info,e1,e2,e3)) = 
    let
	val (b,t1,t2) = (check' g e1, check' g e2, check' g e3) 
	val sup_type  = sup (t1,t2)
    in	
	(if compare g (b,BoolType emptyInfo) then 
	     if (!subtyping orelse compare g (t1,t2)) then 
		 ((!squiggle) (Star emptyInfo) (check' g sup_type);sup_type)
	     else raise ErrorWInfo (info,
				    [Str "types of conclusion and alternative",
				     Str " of if must equivalent",
				     Str " but ", E t1, Str " <> ", E t2, 
				     Str " in the current mode"])
	 else raise ErrorWInfo (info,
				[Str "expected type Bool for if head but got ",
				 E b]))
    end
  | checkIf _ _ = 
    raise Error "internal error: wrong function application for check'If"

and checkField g xs info error_msg = 
    let
	fun test (_,e) = ((!squiggle) (Star emptyInfo) (check' g e); ())
    in
	if null xs then raise ErrorWInfo(info,[Str "empty field",
					       Str " type is not allowed"])
	else (app test xs; Star emptyInfo) 
	     handle ErrorWInfo (info,xs) => 
		    raise ErrorWInfo(info,Str (" "^error_msg^"\n! ") :: xs)
    end					

and checkVariant g (Variant(info,s,e1,Box _ )) =
    if !subtyping then VariantSort(emptyInfo,[(s, check' g e1)])
    else raise ErrorWInfo(info,[Str "variant requires an ascription",
				Str "in non-subtyping mode"])
  | checkVariant g (Variant (info,s,e1,e2)) =
    let 
	val A  = check' g e1
	val _  = check' g e2
    in 
	case eval e2 of 
	    VariantSort (info,xs) => 
	    (case List.find (fn (l,_) => s=l) xs of
		 NONE => raise ErrorWInfo (info,
					   [Str "component type ",E A, 
					    Str " does not match ascription ",
					    E e2]) 
	       | SOME (_,T) => 
		     if compare g (A,T) then e2 
		     else raise ErrorWInfo 
			 (info,[Str "component type ",
			   E A, 
			   Str " does not match ascription ",E e2]) )
	  | _ => raise ErrorWInfo (info,
				   [Str "wrong variant ascription ", E e2])    
    end
  | checkVariant _ _ = 
    raise Error "internal error: wrong function application for check'Variant"

 and checkCase g (Case(info,e,xs)) =
     let	
	 (* get the type for e first *)
	 val ety_list = 
	     (case eval (check' g e) of 
		  VariantSort (_,xs) => xs
		| T => raise ErrorWInfo (info,
					 [Str "expected variant value,",
					  Str "but got value of type ", E T]))

	 fun find_label_type l xs = 
	     (case List.find (fn (l',_) => l=l') ety_list of
		  NONE => raise ErrorWInfo(info,[Str "variant type does",
						 Str "not correspond ",
						 Str "to case patterns"])
		| SOME (_,T) => T)

	 fun check_component ((l,x,e),T) = 
	     let
		 val T' = check' (find_label_type l xs :: g) e
		 val sup_type = sup(T,T')
	     in
		 if (!subtyping orelse compare g (T',T)) then 
		     ((!squiggle) 
		      (Star emptyInfo) 
		      (check' g sup_type);
		      sup_type) 
		 else raise ErrorWInfo (info,
					[Str "expected type ", E T, 
					 Str " in case component, but got ",
					 E T'])
	     end
					   
	 val T1 = case (hd xs) of 
	     (l,x,e) => check' (find_label_type l xs :: g) e
     in 
	 (!squiggle) (Star emptyInfo) (check' g T1);
	 foldl check_component T1 (tl xs)
    end
   | checkCase _ _ = 
    raise Error "internal error: wrong function application for checkCase"

end
    
