(*-----------------------------------------------------------------------------
   pretty printer for lambda cube syntacs
   
   Call pplcwredex to underline the current available redex.

   author: Christian Mueller
 ----------------------------------------------------------------------------*)

local 
    open AbstractSyntax LambdaCube LCSupport Info
in

(* global attributes *)
val ppconfig = { printMode = ref Opaque, 
		 printOutput = ref All }

val pstream     = { buffer = ref "", 
		    linewidth = ref 80, 
		    redexstart = ref ~1, 
		    redexlen = ref 0, 
		    mark_redex = ref false }

(* temp stream that can be switched with store and restore *)
val tmpstream = { buffer = ref "", 
        	  linewidth = ref 80, 
		  redexstart = ref ~1, 
		  redexlen = ref 0, 
		  mark_redex = ref true }

fun store () = (#buffer tmpstream     := !(#buffer pstream);
		#linewidth tmpstream  := !(#linewidth pstream);
		#redexstart tmpstream := !(#redexstart pstream);
		#redexlen tmpstream   := !(#redexlen pstream);
		#mark_redex tmpstream := !(#mark_redex pstream))

fun restore () = (#buffer pstream     := !(#buffer tmpstream);
		  #linewidth pstream  := !(#linewidth tmpstream);
		  #redexstart pstream := !(#redexstart tmpstream);
		  #redexlen pstream   := !(#redexlen tmpstream);
		  #mark_redex pstream := !(#mark_redex tmpstream);
		  #buffer tmpstream     := "";
		  #linewidth tmpstream  := 80;
		  #redexstart tmpstream := ~1;
		  #redexlen tmpstream   := 0;
		  #mark_redex tmpstream := true)

(* functions to modify the behaviour of the pretty printer *)
fun setPrintMode mode     = (#printMode ppconfig) := mode;

fun setPrintOutput output = (#printOutput ppconfig) := output

fun getPrintOutput ()     = !(#printOutput ppconfig)

fun setLineWidth n        = 
    if n < 10 then raise Fail "your linewidth makes no sense" 
    else #linewidth pstream := n

(* functions that are needed for nice printing *)

fun pr_atom (App (info,e1,e2))        = test info
  | pr_atom (Abstr (info,_,_,_))      = test info
  | pr_atom (Pi (info,v,e1,e2))       = test info 
  | pr_atom (Plus(info,e1,e2))        = test info
  | pr_atom (Minus (info,e1,e2))      = test info
  | pr_atom (Mult (info,e1,e2))       = test info 
  | pr_atom (IntEqual (info,e1,e2))   = test info 
  | pr_atom (IntNotEqual (info,e1,e2))= test info
  | pr_atom (Less (info,e1,e2))       = test info 
  | pr_atom (LessEq (info,e1,e2))     = test info 
  | pr_atom (Greater (info,e1,e2))    = test info
  | pr_atom (GreaterEq (info,e1,e2))  = test info 
  | pr_atom (BoolNeg (info,e))        = test info
  | pr_atom (If(info,e1,e2,e3))       = test info 
  | pr_atom (Variant(info,l,e1,e2))   = test info 
  | pr_atom (Case(info,e1,xs))        = test info 
  | pr_atom (Record (info,xs))        = test info 
  | pr_atom (Proj (info,e,l))         = test info 
  | pr_atom (RecursiveSort(info,v,e)) = test info 
  | pr_atom (Fold(info,e1,e2))        = test info 
  | pr_atom (Unfold(info,e1,e2))      = test info 
  | pr_atom (Ascribe(info,e1,e2))     = test info
  | pr_atom (Fix(info,_))             = test info
  | pr_atom (Let(info,_,_,_))         = test info
  | pr_atom _                         = true

(* checks the interpreter mode and whether information is avialable *)
and test info = (!(#printMode ppconfig) <> Transparent) andalso hasId info

fun iskind (Star _)               = true
  | iskind (Pi(_,_,k1,k2))        = iskind k1 andalso iskind k2
  | iskind (RecordSort(_,xs))     = List.all (fn (_,e) => iskind e) xs
  | iskind (VariantSort(_,xs))    = List.all (fn (_,e) => iskind e) xs
  | iskind (RecursiveSort(_,_,e)) = iskind e
  | iskind _                      = false

(* functions to modify the global attribute pstream *)

fun ppredex () = 
    (* prints out a redex if mark_redex is set and pp has marked a redex *)
    let 
	fun f n c s = if n<= 0 then s else f (n-1) c (c^s)
			 
	val x   = (! (#redexstart pstream))

	val len = (! (#redexlen pstream))
    in
	if !(#mark_redex pstream) andalso x <> ~1 then 
	    (let
		 val (x',len') = 
		     if len=0 then (5,String.size(!(#buffer pstream)) - x) 
		     else ( (#mark_redex pstream) := false; (~1,len))
	     in
		 #redexstart pstream := x'; 
		 #redexlen pstream := 0; 
		 print("\n"^f x " " ""^f len' "^" "" )
	     end)
	else ()
    end

fun ppflush () = 
    (* prints out the buffer content and clears the buffer; 
     can be seen from outside *)
    let 
	val s   = (! (#buffer pstream))
    in  
	if s <> "" then (print s; ppredex (); #buffer pstream := "") else (); 
	print "\n";
	#redexstart pstream := ~1;
	#redexlen pstream := 0
    end

fun add s =  
    (* inserts a string into the buffer and breaks if overflow occurs *)
    let 
	val s' = ! (#buffer pstream)
    in 
	if String.size (s' ^ s) > (!(#linewidth pstream)) 
	    then (ppflush (); #buffer pstream := "     "^s)
	else #buffer pstream := s' ^ s
    end
	
(* functions to mark the begin and end of a redex *)
fun begin_redex b = 
    if b then #redexstart pstream := String.size(! (#buffer pstream)) else ()

fun end_redex   b = 
    if b then #redexlen pstream := String.size(!(#buffer pstream))-(!(#redexstart pstream)) 
    else ()

(* functions that can be accessed from outside *)

(* function that fills the buffer with a string created from an expression 
   and marks if a redex is found type applications are not marked *)

fun pplc env exp =
let 
    val rf = ref false (* stores whether a redex has been found *)

    fun setrf b = if b then rf := b else ()

    fun hasdef e = 
	if !(#printMode ppconfig) = OpaqueWithSync then 
	    (case Environment.find (env,e) of 
		 NONE => pr_atom e 
	       | SOME _ => true) 
	else pr_atom e		    

    fun padd info f = if !(#printMode ppconfig) <> Transparent 
			 andalso hasId info 
			  then 
			      (let val rf' = !rf 
				   val s = getId info
			       in
				   store ();			       
				   f ();
				   restore ();
				   begin_redex (not rf' andalso (!rf));
				   add s;
				   end_redex (not rf' andalso (!rf))
			       end)
		      else f ()

    (* search first in env for a declaration *)
    fun pp e = if !(#printMode ppconfig) = OpaqueWithSync then
		   ( case Environment.find (env,e) of 
			 NONE => pp' e 
		       | SOME s => add s )
	       else pp' e

    (* now build the string corresponding to the expression *)
    and pp' (Star info) = padd info (fn () => add "*")
      | pp' (Box info)  = padd info (fn () => add "[]")
      | pp' (Var(info,s,x)) =  padd info (fn () => add s)
      | pp' (App(info,e1,e2)) = 
	if !(#printMode ppconfig) <> Transparent andalso hasId info 
	    then add (getId info)
	else 
	    let 
		val rf'   = not (!rf) andalso isval e1 andalso isval e2
	    in 
		begin_redex rf';
		case e1 of App _ => pp e1 | _ => ppparent e1;
		add " "; 
		ppparent e2;					       
		end_redex rf';
		setrf rf'
	    end

      | pp' (Pi(info,s,e1,e2)) = 
	let
	    val temp = !rf
	in
	    rf := true;
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then add (getId info)
	    else (if s <> "" andalso occurs e2 0 then 			    
		      if equiv (e1, Star emptyInfo) 
		         then (add ("\\/"^s^"."); pp e2)
		      else if iskind e1 
			       then (add ("\\/"^s^":"); pp e1; add "."; pp e2)
			   else (add("("^s^"::");pp e1; add ")->"; pp e2)
		  else	   
		      if hasdef e1 then (pp e1;  add "->"; pp e2)
		      else (add "("; pp e1; add ")->"; pp e2) );	    
	    rf := temp
	end
      | pp' (Abstr(info,s,e1,e2)) = 
	let 
	    val temp = !rf
	in  
	    rf := true;
	    if !(#printMode ppconfig) <> Transparent then 
		if hasId info then add (getId info) else add "<fun>"
	    else (add ("\\"^s^":"); pp e1; add "."; pp e2);
		rf := temp
	end

      | pp' (Int (info,i))        = padd info (fn () => add (Int.toString i))
      | pp' (IntNeg (info,e))         = padd info (fn () => (add "~"; pp e))
      | pp' (Plus (info,e1,e2))       = padd info (fn () => ppbinop e1 e2 "+")
      | pp' (Minus(info,e1,e2))       = padd info (fn () => ppbinop e1 e2 "-")
      | pp' (IntEqual (info,e1,e2))   = padd info (fn () => ppbinop e1 e2 "=")
      | pp' (IntNotEqual (info,e1,e2))= padd info (fn () => ppbinop e1 e2 "<>")
      | pp' (LessEq (info,e1,e2))     = padd info (fn () => ppbinop e1 e2 "<=")
      | pp' (Less (info,e1,e2))       = padd info (fn () => ppbinop e1 e2 "<")
      | pp' (GreaterEq (info,e1,e2))  = padd info (fn () => ppbinop e1 e2 ">=")
      | pp' (Greater (info,e1,e2))    = padd info (fn () => ppbinop e1 e2 ">")
      | pp' (Mult (info,e1,e2))       = 
	padd info (fn () => 
		   let (* set parenthesis where needed *)
		       val rf' = not (!rf) andalso isval e1 andalso isval e2 
		   in
		       begin_redex rf';
		       case e1 of 
			   Plus  (info,_,_) => 
			   if hasId info then pp e1 
			   else (add "("; pp e1; add ")") 

			 | Minus (info,_,_) => 
			   if hasId info then pp e1 
			   else (add "("; pp e1; add ")")

			 | _ => pp e1;
			 			
		       add "**";
		       case e2 of 
			   Plus  (info,_,_) => 
			   if hasId info then pp e2
			   else (add "("; pp e2; add ")") 

			 | Minus (info,_,_) => 
			   if hasId info then pp e1 
			   else (add "("; pp e2; add ")")

			 | _ => pp e2;

		       end_redex rf';
		       setrf rf'
		   end)					     
      | pp' (IntType info)    = padd info (fn () => add "Int")
      | pp' (BoolType info)   = padd info (fn () => add "Bool")		   
      | pp' (True info)       = padd info (fn () => add "true")
      | pp' (False info)      = padd info (fn () => add "false")
      | pp' (BoolNeg (info,e))  = padd info (fn () => (add "not "; pp e))
      | pp' (If(info,e1,e2,e3)) = 
	let 
	    val rf' = not (!rf) andalso isval e1			       
	in 
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then (begin_redex rf'; add (getId info); end_redex rf';
		      setrf rf')
	    else (begin_redex rf';
		  add "if ";  pp e1; add " then "; 
		  (let
		       val temp = !rf 
		   in 
		       rf := true;pp e2; add " else "; pp e3; 
		       rf := temp 
		   end);
		   end_redex rf';
		   setrf rf')
	end
      | pp' (Variant(info, l,e,Box _)) = 
	padd info (fn () => (add ("<"^l^"="); pp e; add ">"))
	
      | pp' (Variant(info, l,e,T)) = 
	padd info (fn () => (add ("<"^l^"="); pp e; add "> as "; pp T))
      | pp' (VariantSort (info,xs)) = 
	let 
	    val temp = !rf
	in 
	    rf := true;
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then add (getId info)
	    else (add"<"; pplst xs ":" ; add ">");
	    rf := temp
	end

      | pp' (Case(info,e,xs)) = 
	let 				
	    fun prCase nil = ()
              | prCase ((l,x,e)::xr) = (add ("| <"^l^"="^x^"> => "); 
					pp e; prCase xr)
				       
	    val rf' = not (!rf) andalso isval e
	in
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
	    then (begin_redex rf'; add (getId info); end_redex rf'; setrf rf')
	    else (begin_redex rf'; add "case "; pp e; add " of "; 
		  case xs of ((l,x,e)::xr)=>(add (" <"^l^"="^x^"> => "); 
					     pp e; prCase xr)| _ => ();
		  end_redex rf';
		  setrf rf')				      
        end

      | pp' (Proj(info,e,s)) = 
	let 
	    val rf'  = not (!rf) andalso isval e			       
	in
	    begin_redex rf'; 
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then (add (getId info); end_redex rf'; setrf rf') 
	    else (pp e; add ("@"^s); end_redex rf'; setrf rf')
	end				

      | pp' (Record(info,xs)) = 
	if !(#printMode ppconfig) <> Transparent andalso hasId info 
	    then add (getId info)
	else (add "{"; pplst xs "="; add "}")

      | pp' (RecordSort(info,xs))    = 
	let 
	    val temp = !rf
	in
	    rf := true;
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then (add (getId info))
	    else (add "{"; pplst xs ":"; add "}");
	    rf := temp
	end

      | pp' (Fold(info,T,e)) = 
	padd info (fn () => (add "fold ";  
			     let 
				 val temp = !rf 
			     in 
				 rf:=true; 
				 ppparent T; rf := temp end; 
				 add " ";
				 ppparent e))

      | pp' (Unfold(info,T,e))            = 
	padd info (fn () => 
		   let
		       val rf' = case e of Fold _ => not (!rf) | _ => false
		   in
		       begin_redex rf';
		       add "unfold "; 
		       let 
			   val temp = !rf 
		       in 
			   rf := true; 
			   ppparent T; rf := temp 
		       end; 
		       add " ";
		       ppparent e;
		       end_redex rf';	    
		       setrf rf'
		   end)

      | pp' (RecursiveSort(info,s,T))= 
	let 
	    val temp = !rf
	in
	    rf := true;					 
	    if !(#printMode ppconfig) <> Transparent andalso hasId info 
		then add (getId info)
	    else (add ("mu "^s^"."); pp T);
	    rf := temp
	end

      | pp' (Unit info)     = padd info (fn () => add "unit")
      | pp' (UnitType info) = padd info (fn () => add "Unit")
      | pp' (Ascribe(info,e,T)) = 
	padd info (fn () =>  (pp e; add " as "; pp T))

      | pp' (Fix(info,e)) = padd info (fn () => (add "fix "; pp e))

      | pp' (Ref(info,e)) = padd info (fn () => (add "ref "; pp e))
      | pp' (Deref(info,e)) = padd info (fn () => (add "!"; pp e))
      | pp' (RefType(info,e)) = padd info (fn () => (add "Ref "; pp e))
      | pp' (l as Loc(info,x)) = 
(*	padd info (fn () => add "loc")*)
	padd info (fn () => (add "ref "; pp (eval (Deref(emptyInfo,l)))))      

      | pp' (Assign(info,e1,e2)) = 
	padd info (fn () => (pp e1;add " := ";pp e2))

      | pp' (Let(info,s,e1,e2)) =
	padd info 
	(fn () => 
	    let 
		val rf'   = not (!rf) andalso isval e1
	    in 
		begin_redex rf';
		add ("let "^s^"=");
		pp e1;		
		add " in "; 
		pp e2;					       
		end_redex rf';
		setrf rf'
	    end)
		      
    and ppbinop e1 e2 sop = 
	let 
	    val rf'  = not (!rf) andalso isval e1 andalso isval e2 
	in  
	    begin_redex rf'; 
	    pp' e1; 
	    add sop; 
	    case e2 of Mult _ => pp' e2 |  _ => ppparent e2;
		end_redex rf';
		setrf rf'
	end
				
    and ppparent e = if pr_atom e then pp e else (add "("; pp' e; add ")")
						 
    and pplst xs sep = 
	let
	    fun f nil         = ()
	      | f [(l,e)]     = (add (l^sep); pp' e) 
	      | f ((l,e)::xr) = (add (l^sep); pp' e; add ","; f xr)
	in 
	    f xs
	end	    
in
    pp exp
end

(* fills the buffer with a string *)
fun pprint s = add s 

(* outputs a string without using context information *)

val empty_env = Environment.create ()

fun pp e = pplc empty_env e
    
(* prints an expression with its redex marked and the corresponding type *)
fun pplcwredex env e T = 
    ( (#mark_redex pstream):=true; 
      if getPrintOutput () = All orelse getPrintOutput () = Term 
	  then pplc env e else pprint "<term>";
      if getPrintOutput () = All orelse getPrintOutput () = Type 
	  then (pprint " : "; pplc env T) else ();
      ppflush ();
      (#mark_redex pstream):=false )

end 
